#include "core.h"
#include "hashedstring.h"
#include "rawtexture.h"
#include "filestream.h"
#include "lodepng.h"

#include <string.h> 

RawTexture::RawTexture()
: m_ARGBImage( 0 )
{
}

/*virtual*/ RawTexture::~RawTexture()
{
	if( m_ARGBImage != NULL )
	{
		delete[] m_ARGBImage;
	}
}

/*virtual*/ void* RawTexture::GetHandle()
{
	return m_ARGBImage;
}

void RawTexture::Initialize( uint Width, uint Height )
{
	m_Width = Width;
	m_Height = Height;
	m_ARGBImage = new byte[ Width * Height * 4 ];
	memset( m_ARGBImage, 0, Width * Height * 4 );
}

void RawTexture::SavePNG( const IDataStream& Stream, int Width, int Height, byte* ARGBImage )
{
	byte *PNGImage = new byte[ Width * Height * 4 ];
	byte *From = ARGBImage, *To = PNGImage;
	byte R, G, B, A;
	for ( int Y = 0; Y < Height; ++Y )
	{
		for ( int X = 0; X < Width; ++X )
		{
			B = *From++;
			G = *From++;
			R = *From++;
			A = *From++;

			*To++ = R;
			*To++ = G;
			*To++ = B;
			*To++ = A;
		}
	}

	unsigned char *Result = NULL;
	size_t ResultSize = 0;
	if( lodepng_encode32( &Result, &ResultSize, PNGImage, Width, Height ) == 0 )
	{
		Stream.Write( ResultSize, Result );
	}
	else
	{
		WARNDESC( "SavePNG: image encoding failed." );
	}
	delete[] PNGImage;
	if( Result )
	{
		free( Result );
	}
}

void RawTexture::Save( const char* const Filename )
{
	Save( FileStream( Filename, FileStream::EFM_Write ) );
}

void RawTexture::Save( const IDataStream &Stream )
{
	SavePNG( Stream, m_Width,  m_Height, m_ARGBImage );
}

void RawTexture::CopyRect( RawTexture &From, uint FromX, uint FromY, uint Width, uint Height, uint ToX, uint ToY )
{
	if ( this == &From )
	{
		WARNDESC( "Copyrect: can't copy to the same image you copy from" );
		return;
	}

	if ( ToX + Width > m_Width || ToY + Height > m_Height || FromX + Width > From.m_Width || FromY + Height > From.m_Height )
	{
		WARNDESC("Copyrect: coordinates out of bounds");
		// TODO might handle this gracefully
		return;
	}
	uint ToLL = m_Width * 4;
	byte *pTo = m_ARGBImage + ToY * ToLL + ToX * 4;
	uint ToSkip = ToLL - Width * 4;
	uint FromLL = From.m_Width * 4;
	byte *pFrom = From.m_ARGBImage + FromY * FromLL + FromX * 4;
	uint FromSkip = FromLL - Width * 4;

	for( uint y = 0; y < Height; ++y )
	{
		for( uint x = 0; x < Width; ++x )
		{
			*pTo++ = *pFrom++; // A
			*pTo++ = *pFrom++; // R
			*pTo++ = *pFrom++; // G
			*pTo++ = *pFrom++; // B
		}
		pTo += ToSkip;
		pFrom += FromSkip;
	}
}

void RawTexture::SetPixel( uint X, uint Y, uint8 R, uint8 G, uint8 B, uint8 A )
{
	if( X >= m_Width || Y >= m_Height )
	{
		return;
	}
	uint8 *pPixel = m_ARGBImage + Y * m_Width * 4 + X * 4;
	*pPixel++ = B;
	*pPixel++ = G;
	*pPixel++ = R;
	*pPixel = A;
}

bool RawTexture::GetPixel( uint X, uint Y, uint8 &R, uint8 &G, uint8 &B, uint8 &A )
{
	if( X >= m_Width || Y >= m_Height )
	{
		return false;
	}
	uint8 *pPixel = m_ARGBImage + Y * m_Width * 4 + X * 4;
	B = *pPixel++;
	G = *pPixel++;
	R = *pPixel++;
	A = *pPixel;
	return true;
}

/* virtual */ void	RawTexture::CreateTexture( byte* const ARGBImage )
{
	// should only be used internally by TextureCommon::Initalize(),
	// so storing the pointer should be safe
	m_ARGBImage = ARGBImage;
}

/* virtual */ void	RawTexture::CreateTextureFromDDS( const IDataStream& Stream )
{
	WARNDESC("RawTexture::CreateTextureFromDDS not supported");
	m_Width = 0;
	m_Height = 0;
	m_ARGBImage = NULL;
}
