#include "core.h"
#include "array.h"
#include "map.h"
#include "simplestring.h"
#include "cmdlineparser.h"
#include <stdlib.h>
#include <stdio.h>

CmdlineParser::CmdlineParser( int Argc, char *Argv[] )
: m_IsParsed( false )
, m_Argv()
, m_MinPositionalArgs( 0 )
, m_MaxPositionalArgs( 0 )
, m_PositionalNames( "" )
, m_PositionalDescription( "" )
, m_ProgramName( "" )
, m_Options()
, m_Values()
, m_PositionalValues()
{
	if( Argc <= 0 || Argv == NULL )
	{
		Error( "Invalid Argc/Argv Parameters to CmdlineParser" );
	}
	m_ProgramName = Argv[0];

	for( int i = 1; i < Argc; ++i )
	{
		m_Argv.PushBack( Argv[i] );
	}
}

CmdlineParser::~CmdlineParser()
{
}

bool CmdlineParser::AddOption( int Option, const char *Name, const char *Description, bool WantArg, const char *Default )
{
	if ( Name && ( Option > ' ' && Option < 127 ) && Description )
	{
		m_Options.PushBack( CmdlineParser::OptionDesc( Option, Name, Description, WantArg, Default ) );
		return true;
	}
	return Error( "Invalid parameters to AddOption" );
}

bool CmdlineParser::AddPositionals( unsigned int Min, unsigned int Max, const char *Names, const char *Description )
{
	if ( ( Min > 0 && !Names ) || ( Max < Min && Max > 0 ) || !Description )
	{
		return Error( "Invalid parameters to AddPositionals" );;
	}
	m_MinPositionalArgs = Min;
	m_MaxPositionalArgs = Max;
	m_PositionalNames = Names;
	m_PositionalDescription = Description;
	return true;
}

bool CmdlineParser::AddPositionals( unsigned int Num, const char *Names, const char *Description )
{
	return AddPositionals( Num, Num, Names, Description );
}

void CmdlineParser::SetDefaultValues()
{
	FOR_EACH_ARRAY( Iter, m_Options, OptionDesc )
	{
		OptionDesc &Opt = Iter.GetValue();
		if ( Opt.m_HasDefault )
		{
			m_Values[Opt.m_Name.CStr()] = Opt.m_Default;
		}
	}
}

CmdlineParser::OptionDesc *CmdlineParser::FindOption( SimpleString Option )
{
	FOR_EACH_ARRAY( Iter, m_Options, OptionDesc )
	{
		OptionDesc &Opt = Iter.GetValue();
		if( Option == Opt.m_Option )
		{
			return &Opt;
		}
	}
	return NULL;
}

bool CmdlineParser::Parse()
{
	SetDefaultValues();

	Array<SimpleString>::Iterator ArgIter = m_Argv.Begin();
	while( ArgIter.IsValid() )
	{
		SimpleString &Arg = ArgIter.GetValue();
		if( *Arg.CStr() == '-' )
		{
			SimpleString Opt, Val;
			if( Arg.Length() == 2 )
			{
				Opt = Arg;
			}
			else if( Arg.Length() > 2 )
			{
				Arg.Split( 2, Opt, Val );
			}
			else
			{
				return false;
			}
			OptionDesc *pOpt = FindOption( Opt );
			if( !pOpt )
			{
				return false;
			}
			if( pOpt->m_WantArg )
			{
				if(Val == "" )
				{
					ArgIter++;
					Val = ArgIter.GetValue(); // TODO check for error
				}
				m_Values[ pOpt->m_Name ] = Val;
			}
			else if( Val == "" )
			{
				m_Values[ pOpt->m_Name ] = "true";
			}
			else
			{
				return false;
			}
		}
		else
		{
			m_PositionalValues.PushBack( Arg );
		}
		ArgIter++;
	}

	if( m_PositionalValues.Size() < m_MinPositionalArgs ||
		( m_MaxPositionalArgs >= 0 && m_PositionalValues.Size() > m_MaxPositionalArgs ) )
	{
		return false;
	}
	
	m_IsParsed = true;
	return true;
}

SimpleString CmdlineParser::Usage()
{
	SimpleString Result = SimpleString::PrintF( "Syntax: %s", m_ProgramName.CStr() );
	bool HasPositionalArgs = m_MinPositionalArgs > 0 || m_MaxPositionalArgs > 0;
	FOR_EACH_ARRAY( Item, m_Options, OptionDesc )
	{
		OptionDesc &Opt = Item.GetValue();
		if( Opt.m_WantArg )
		{
			Result += SimpleString::PrintF( " [%s %s]", Opt.m_Option.CStr(), Opt.m_Name.CStr() );
		}
		else
		{
			Result += SimpleString::PrintF( " [%s]", Opt.m_Option.CStr() );
		}
	}
	if ( HasPositionalArgs )
	{
		Result += SimpleString::PrintF( " %s", m_PositionalNames.CStr() );
	}

	FOR_EACH_ARRAY( Item, m_Options, OptionDesc )
	{
		OptionDesc &Opt = Item.GetValue();
		if( Opt.m_WantArg )
		{
			Result += SimpleString::PrintF( "\n  %s %s", Opt.m_Option.CStr(), Opt.m_Name.CStr() );
		}
		else
		{
			Result += SimpleString::PrintF( "\n  %s", Opt.m_Option.CStr() );
		}
		Result += SimpleString::PrintF( "\n    %s", Opt.m_Description.CStr() );
		if( Opt.m_HasDefault )
		{
			Result += SimpleString::PrintF( "\n    Default: %s", Opt.m_Default.CStr() );
		}
	}
	if ( HasPositionalArgs )
	{
		Result += SimpleString::PrintF( "\n  %s", m_PositionalNames.CStr() );
		Result += SimpleString::PrintF( "\n    %s", m_PositionalDescription.CStr() );
	}
	Result += "\n";

	return Result;
}

const char *CmdlineParser::operator[]( const char *Name )
{
	if( !m_IsParsed )
	{
		Error( "CmdlineParser has not yet parsed the command line!" );
		return NULL;
	}
	Map<SimpleString, SimpleString>::Iterator Item = m_Values.Search( Name );
	if( Item.IsValid() )
	{
		return Item.GetValue().CStr();
	}
	return NULL;
}

const char *CmdlineParser::operator[]( int Position )
{
	if( !m_IsParsed )
	{
		Error( "CmdlineParser has not yet parsed the command line!" );
		return NULL;
	}
	if( Position > 0 && (unsigned int)Position <= m_PositionalValues.Size() )
	{
		return m_PositionalValues[ Position - 1 ].CStr();
	}
	return NULL;
}

const char *CmdlineParser::ProgramName()
{
	return m_ProgramName.CStr();
}

bool CmdlineParser::Error( const char *Message )
{
	fprintf( stderr, "%s\n", Message );
	BREAKPOINT;
	return false;
}
