#include "core.h"
#include "customassert.h"
#include "simplestring.h"
#include "printmanager.h"
#include "configmanager.h"
#include "luainterpreter.h"

#include <stdarg.h>
#include <string.h>
#include <stdio.h>

LuaInterpreter* LuaInterpreter::m_Instance = NULL;

// pcall error handler from lua.c
static int PcallErrorHandler( lua_State *L )
{
  const char *msg = lua_tostring(L, 1);
  if (msg == NULL) {  /* is error object not a string? */
    if (luaL_callmeta(L, 1, "__tostring") &&  /* does it have a metamethod */
        lua_type(L, -1) == LUA_TSTRING)  /* that produces a string? */
      return 1;  /* that is the message */
    else
      msg = lua_pushfstring(L, "(error object is a %s value)",
                               luaL_typename(L, 1));
  }
  luaL_traceback(L, L, msg, 1);  /* append a standard traceback */
  return 1;  /* return the traceback */
}

static int BaseLog( lua_State *L )
{
	if ( lua_gettop( L ) > 1 )
	{
		int nargs = lua_gettop( L );
		lua_getglobal( L, "string" );
		lua_getfield( L, -1, "format" );
		lua_insert( L, 1 );
		lua_pop( L, 1 );
		lua_call( L, nargs, 1 );
	}

	const char *Msg = luaL_checkstring(L, 1);
	PRINTF( "Lua: %s\n", Msg );
	return 0;
}

LuaInterpreter *LuaInterpreter::GetInstance()
{
	if( !m_Instance )
	{
		m_Instance = new LuaInterpreter;
	}
	return m_Instance;
}

void LuaInterpreter::DeleteInstance()
{
	SafeDelete( m_Instance );
}

LuaInterpreter::LuaInterpreter()
: m_LuaState( NULL )
, m_LastError( NULL )
{
	m_LuaState = luaL_newstate();
	luaL_openlibs( m_LuaState ); // TODO custom?

	lua_newtable( m_LuaState );
	lua_setfield( m_LuaState, LUA_REGISTRYINDEX, "CONTEXT_BY_HASH" );

	lua_newtable( m_LuaState );
	lua_setfield( m_LuaState, LUA_REGISTRYINDEX, "PATH_STACK" );

	SetString( "cpath", "", "package" );
	SetString( "path", "?.lua;Lua/?.lua", "package" );

	AddCFunction( "log", BaseLog );
}

LuaInterpreter::~LuaInterpreter()
{
	if( m_LuaState )
	{
		lua_close( m_LuaState );
	}
	if ( m_LastError )
	{
		delete[] m_LastError;
	}
}

void LuaInterpreter::PushPath( const char *Path )
{
	lua_getfield( m_LuaState, LUA_REGISTRYINDEX, "PATH_STACK" );
	size_t Pos = lua_rawlen( m_LuaState, -1 );
	lua_getglobal( m_LuaState, "package" );
	lua_getfield( m_LuaState, -1, "path");
	const char *PrevPath = lua_tostring( m_LuaState, -1 );
	lua_rawseti( m_LuaState, -3, Pos + 1 );
	lua_pushfstring( m_LuaState, "%s/?.lua;%s", Path, PrevPath );
	lua_setfield( m_LuaState, -2, "path" );
	lua_pop( m_LuaState, 2 );
}

void LuaInterpreter::PopPath()
{
	lua_getfield( m_LuaState, LUA_REGISTRYINDEX, "PATH_STACK" );
	size_t Pos = lua_rawlen( m_LuaState, -1 );
	lua_getglobal( m_LuaState, "package" );
	lua_rawgeti( m_LuaState, -2, Pos );
	lua_setfield( m_LuaState, -2, "path" );
	lua_pop( m_LuaState, 1 );
	lua_pushnil( m_LuaState );
	lua_rawseti( m_LuaState, -2, Pos );
	lua_pop( m_LuaState, 1 );
}

int LuaInterpreter::Error( const char *Msg, ... )
{
	ClearError();
	va_list Arglist;
	va_start( Arglist, Msg );
	int len = vsnprintf( NULL, 0, Msg, Arglist ) + 1;
	va_end( Arglist );
	m_LastError = new char[ len ];
	va_start( Arglist, Msg );
	vsnprintf( m_LastError, len, Msg, Arglist );
	va_end( Arglist);
	return 0;
}

void LuaInterpreter::ClearError()
{
	if( m_LastError )
	{
		delete[] m_LastError;
	}
	m_LastError = NULL;
}

bool LuaInterpreter::Pcall( int Nargs, int Nrets )
{
	int HandlerPos = lua_gettop( m_LuaState ) - Nargs;
	lua_pushcfunction( m_LuaState, PcallErrorHandler );
	lua_insert( m_LuaState, HandlerPos );
	int Res = lua_pcall( m_LuaState, Nargs, Nrets, HandlerPos );
	lua_remove( m_LuaState, HandlerPos );
	if ( Res != LUA_OK )
	{
		Error( "%s", lua_tostring( m_LuaState, -1 ) );
		lua_pop( m_LuaState, 1 );
		return false;
	}
	return true;
}

int LuaInterpreter::GetContext( const char *CtxName )
{
	int CType = lua_getglobal( m_LuaState, CtxName );
	if ( CType == LUA_TNIL )
	{
		lua_pop( m_LuaState, 1 );
		lua_newtable( m_LuaState );

		// set metatable
		lua_newtable( m_LuaState );
		lua_getglobal( m_LuaState, "_G" );
		lua_setfield( m_LuaState, -2, "__index" );
		lua_setmetatable( m_LuaState, -2 );

		// set context name
		lua_pushstring( m_LuaState, CtxName );
		lua_setfield( m_LuaState, -2, "CONTEXT" );

		// set context by name
		lua_pushvalue( m_LuaState, -1 );
		lua_setglobal( m_LuaState, CtxName );

		// set context by hash
		lua_getfield( m_LuaState, LUA_REGISTRYINDEX, "CONTEXT_BY_HASH" );
		lua_pushvalue( m_LuaState, -2 );
		lua_seti( m_LuaState, -2, HashedString::Hash( CtxName ) );
		lua_pop( m_LuaState, 1 );

	}
	else if( CType != LUA_TTABLE )
	{
		lua_pop( m_LuaState, 1 );
		return Error( "invalid context: %s", CtxName );
		return 0;
	}
	return 1;
}

int LuaInterpreter::GetContextByHash( const HashedString& Context )
{
	lua_getfield( m_LuaState, LUA_REGISTRYINDEX, "CONTEXT_BY_HASH" );
	if ( lua_geti( m_LuaState, -1, Context.GetHash() ) != LUA_TTABLE )
	{
		lua_pop( m_LuaState, 2 );
		return 0;
	}
	lua_remove( m_LuaState, -2 );
	return 1;
}

int LuaInterpreter::Eval( const char *Data, const char *Name, bool ResetStack )
{
	if( ResetStack )
	{
		lua_settop( m_LuaState, 0 );
	}
	int Top = lua_gettop( m_LuaState );
	ClearError();

	if ( luaL_loadbufferx( m_LuaState, Data, strlen( Data ), Name, "t" ) != LUA_OK )
	{
		Error( "%s", lua_tostring( m_LuaState, -1 ) );
		lua_pop( m_LuaState, 1 );
		return 0;
	}

	if( !Pcall( 0, LUA_MULTRET ) )
	{
		return 0;
	}

	return lua_gettop( m_LuaState ) - Top;
}

void LuaInterpreter::PrintStack()
{
	int Num = lua_gettop( m_LuaState );
	if ( Num > 0 )
	{
		lua_getglobal( m_LuaState, "print" );
		lua_insert( m_LuaState, 1 );
		if ( !Pcall( Num, 0 ) )
		{
			printf("Error while dumping stack: %s\n", LastError() );
		}
	}
}

int LuaInterpreter::DoEvalInContext( const char *Data, const char *Context, const char *Name, int NResults )
{
	lua_settop( m_LuaState, 0 );
	ClearError();

	if ( luaL_loadbufferx( m_LuaState, Data, strlen( Data ), Name, "t" ) != LUA_OK )
	{
		Error( "%s", lua_tostring( m_LuaState, -1 ) );
		lua_pop( m_LuaState, 1 );
		return LUA_TNONE;
	}

	if( Context && *Context )
	{
		if ( !GetContext( Context ) )
		{
			return LUA_TNONE;
		}

		if ( !lua_setupvalue( m_LuaState, -2, 1 ) )
		{
			Error( "Could not set context: %s", Context );
			lua_pop( m_LuaState, 2 );
			return LUA_TNONE;
		}
	}

	if( !Pcall( 0, NResults ) )
	{
		return  LUA_TNONE;
	}

	return NResults > 0 ? lua_type( m_LuaState, -1 ) : LUA_TNIL;
}

bool LuaInterpreter::EvalInContext( const char *Data, const char *Context, const char *Name )
{
	return DoEvalInContext( Data, Context, Name, 1 ) > LUA_TNONE;
}

bool LuaInterpreter::RunInContext( const char *Data, const char *Context, const char *Name )
{
	DoEvalInContext( Data, Context, Name, 0 );
	return m_LastError == NULL;
}

// TODO handle errors
bool LuaInterpreter::CallGetvarHandler( const HashedString &Var, const HashedString &Context )
{
	ClearError();
	if ( Context != ConfigManager::EmptyContext )
	{
		if ( GetContextByHash( Context ) )
		{
			if ( lua_getfield( m_LuaState, -1, "GETVAR" ) == LUA_TFUNCTION )
			{
				lua_pushinteger( m_LuaState, Var.GetHash() );
				lua_pushinteger( m_LuaState, Context.GetHash() );
				if ( !Pcall( 2, 1 ) )
				{
					return false;
				}
				if ( !lua_isnil( m_LuaState, -1 ) )
				{
					return true;
				}
			}
			lua_pop( m_LuaState, 2 );
		}
	}
	if ( lua_getglobal( m_LuaState, "GETVAR" ) == LUA_TFUNCTION )
	{
		lua_pushinteger( m_LuaState, Var.GetHash() );
		lua_pushinteger( m_LuaState, Context.GetHash() );
		return Pcall( 2, 1 );
	}
	lua_pop( m_LuaState, 1 );
	return false;
}

bool LuaInterpreter::RunFile( const char *Filename )
{
	ClearError();
	lua_settop( m_LuaState, 0 );
	if ( luaL_loadfilex( m_LuaState, Filename, "t" ) != LUA_OK )
	{
		Error( "%s", lua_tostring( m_LuaState, -1 ) );
		lua_pop( m_LuaState, 1 );
		return false;
	}

	if( !Pcall( 0, 0 ) )
	{
		return false;
	}
	return true;
}

bool LuaInterpreter::IsString()
{
	return lua_gettop( m_LuaState ) > 0 &&
		lua_type( m_LuaState, -1 ) == LUA_TSTRING;
}

SimpleString LuaInterpreter::PopString()
{
	SimpleString Res = lua_tostring( m_LuaState, -1 );
	lua_pop( m_LuaState, 1 );
	return Res;
}

bool LuaInterpreter::IsInteger()
{
	return lua_gettop( m_LuaState ) > 0 &&
		lua_type( m_LuaState, -1 ) == LUA_TNUMBER &&
		lua_isinteger( m_LuaState, -1 );
}

int LuaInterpreter::PopInteger()
{
	int32 Res = lua_tointeger( m_LuaState, -1 );
	lua_pop( m_LuaState, 1 );
	return Res;
}

bool LuaInterpreter::IsNumber()
{
	return lua_gettop( m_LuaState ) > 0 &&
		lua_type( m_LuaState, -1 ) == LUA_TNUMBER;
}

double LuaInterpreter::PopNumber()
{
	double Res = lua_tonumber( m_LuaState, -1 );
	lua_pop( m_LuaState, 1 );
	return Res;
}

bool LuaInterpreter::IsBoolean()
{
	return lua_gettop( m_LuaState ) > 0 &&
		lua_type( m_LuaState, -1 ) == LUA_TBOOLEAN;
}

bool LuaInterpreter::PopBoolean()
{
	bool Res = lua_toboolean( m_LuaState, -1 );
	lua_pop( m_LuaState, 1 );
	return Res;
}

void LuaInterpreter::Discard()
{
	lua_pop( m_LuaState, 1 );
}

bool LuaInterpreter::AddContext( const char *Context )
{
	lua_newtable( m_LuaState );
	lua_setglobal( m_LuaState, Context );
	return true; // TODO check for errors
}

bool LuaInterpreter::AddCFunction( const char *Name, lua_CFunction Func, const char *Context )
{
	if( !Context )
	{
		Context = "_G";
	}

	if( lua_getglobal( m_LuaState, Context ) != LUA_TTABLE )
	{
		lua_pop( m_LuaState, 1 );
		return luaL_error( m_LuaState, "invalid context: %s", Context );
	}

	lua_pushcfunction( m_LuaState, Func );
	lua_setfield( m_LuaState, -2, Name );
	lua_pop( m_LuaState, 1 );

	return true; // TODO check for errors
}

void LuaInterpreter::SetVar( const char *Name, const char *Context )
{
	if( Context )
	{
		if ( lua_getglobal( m_LuaState, Context ) != LUA_TNIL )
		{
			lua_insert( m_LuaState, -2 );
			lua_setfield( m_LuaState, -2, Name );
		}
		lua_pop( m_LuaState, 1 );
	}
	else
	{
		lua_setglobal( m_LuaState, Name );
	}
}

void LuaInterpreter::SetNil( const char *Name, const char *Context )
{
	lua_pushnil( m_LuaState );
	SetVar( Name, Context );
}

void LuaInterpreter::SetString( const char *Name, const char *Value, const char *Context )
{
	lua_pushstring( m_LuaState, Value );
	SetVar( Name, Context );
}

void LuaInterpreter::SetInteger( const char *Name, int Value, const char *Context )
{
	lua_pushinteger( m_LuaState, Value );
	SetVar( Name, Context );
}

void LuaInterpreter::SetNumber( const char *Name,  double Value, const char *Context )
{
	lua_pushnumber( m_LuaState, Value );
	SetVar( Name, Context );
}

void LuaInterpreter::SetBoolean( const char *Name, bool Value, const char *Context )
{
	lua_pushboolean( m_LuaState, Value );
	SetVar( Name, Context );
}

int LuaInterpreter::RegisterNamespace( const char *Name, const luaL_Reg *Func )
{
	lua_newtable( m_LuaState );
	luaL_setfuncs( m_LuaState, Func, 0 );
	lua_setglobal( m_LuaState, Name );
	return 1;
}

/*static*/ void LuaInterpreter::DumpStack( lua_State *L )
{
	puts( "LUA STACK DUMP" );
	int max = lua_gettop( L );

	for (int i = 1 ; i <= max ; i++)
	{
		lua_getglobal( L, "tostring" );
		lua_pushvalue( L, i );
		lua_call( L, 1, 1 );

		const char *name = lua_tostring( L, -1 );
		const char *type = luaL_typename( L, i );

		printf( " %3d: %s %s\n", i, type, name );
		lua_pop( L, 1 );
	}
}

/*static*/ int LuaInterpreter::UD__Index( lua_State *L )
{
	if( luaL_getmetafield( L, 1, "__methods" ) != LUA_TNIL )
	{
		lua_pushvalue( L, 2 );
		if( lua_gettable( L, -2 ) != LUA_TNIL )
		{
			return 1;
		}
		lua_pop( L, 2 );
	}
	if( luaL_getmetafield( L, 1, "__realindex" ) != LUA_TNIL )
	{
		lua_insert( L, 1 );
		lua_call( L, 2, 1 );
	}
	else
	{
		lua_pushnil( L );
	}
	return 1;
}

/*static*/ int LuaInterpreter::UD__Gc( lua_State *L )
{
	if( luaL_getmetafield( L, 1, "__realgc" ) != LUA_TNIL )
	{
		lua_insert( L, 1 );
		lua_call( L, 1, 0);
	}
	return 0;
}
