#include "core.h"
#include "simplestring.h"
#include "hashedstring.h"
#include "luainterpreter.h"
#include "lualibraries.h"
#include "configmanager.h"
#include "configvar.h"
#include "idatastream.h"
#include "filestream.h"
#include "packstream.h"
#include "memorystream.h"
#include "modmanager.h"

static int ConfigHash( lua_State *L )
{
	const char *String = luaL_checkstring( L, 1 );
	lua_pushinteger( L, HashedString::Hash( String ) );
	return 1;
}

static int ConfigLoad( lua_State *L )
{
	const char *ConfigFile = luaL_checkstring( L, 1 );
	lua_pushboolean( L, ModManager::GetInstance()->LoadConfigFile( ConfigFile ) );
	return 1;
}

static int ConfigLoadstring( lua_State *L )
{
	size_t Len;
	const char *ConfigString = luaL_checklstring( L, 1, &Len );
	// this  is only for reading, so the following const_cast should be safe
	ConfigManager::Load( MemoryStream( const_cast<char*>( ConfigString ), Len ) );
	return  0;
}

static int ConfigSave( lua_State *L )
{
	const char *ConfigFile = luaL_checkstring( L, 1 );
	ConfigManager::Save( FileStream( ConfigFile, FileStream::EFM_Write ) );
	return  0;
}

static STRING_TYPE VarHash( lua_State *L, int Pos )
{
	if( lua_type( L, Pos ) == LUA_TNUMBER )
	{
		return STRING_TYPE( (uint32)lua_tointeger( L, Pos ) );
	}
	else if( lua_type( L, Pos ) == LUA_TSTRING )
	{
		return STRING_TYPE( lua_tostring( L, Pos ) );
	}
	return luaL_error( L, "invalid argument type, expected string or integer" );
}

static STRING_TYPE ContextHash( lua_State *L, int Pos )
{
	if (!lua_isnoneornil( L, Pos ) )
	{
		return VarHash( L, Pos );
	}
	return ConfigManager::EmptyContext;
}

static int ConfigGetvar( lua_State *L )
{
	STRING_TYPE Name = VarHash( L, 1 );
	STRING_TYPE Context = ContextHash( L, 2 );
	ConfigVar Var = ConfigManager::GetConfigVar( Name, Context );

	switch ( Var.m_Type )
	{
	//case ConfigVar::EVT_None:
	case ConfigVar::EVT_Bool:
		lua_pushboolean( L, Var.m_Bool );
		break;
	case ConfigVar::EVT_Int:
		lua_pushinteger( L, Var.m_Int );
		break;
	case ConfigVar::EVT_Float:
		lua_pushnumber( L, Var.m_Float );
		break;
	case ConfigVar::EVT_String:
		lua_pushstring( L, Var.m_String );
		break;
	default:
		lua_pushnil( L );
	}
	return 1;
}

static int ConfigSetvar( lua_State *L )
{
	STRING_TYPE Name = VarHash( L, 1 );
	STRING_TYPE Context = ContextHash( L, 2 );

	// get the raw var, without reverting to lua if it is not found in the config
	ConfigVar Var = ConfigManager::GetRawConfigVar( Name, Context );

	int Type = lua_type( L, 3 );

	switch( Type )
	{
	case LUA_TBOOLEAN:
		if ( Var.m_Type != ConfigVar::EVT_None && Var.m_Type != ConfigVar::EVT_Bool )
		{
			goto type_error;
		}
		ConfigManager::SetBool( Name, lua_toboolean( L, 3 ), Context );
		break;
	case LUA_TSTRING:
		if ( Var.m_Type != ConfigVar::EVT_None && Var.m_Type != ConfigVar::EVT_String )
		{
			goto type_error;
		}
		ConfigManager::SetString( Name, lua_tostring( L, 3 ), Context );
		break;
	case LUA_TNUMBER:
		if ( Var.m_Type == ConfigVar::EVT_None )
		{
			if ( lua_isinteger( L, 3 ) )
			{
				ConfigManager::SetInt( Name, lua_tointeger( L, 3 ), Context );
			}
			else
			{
				ConfigManager::SetFloat( Name, lua_tonumber( L, 3 ), Context );
			}
		}
		else if( Var.m_Type == ConfigVar::EVT_Int )
		{
			ConfigManager::SetInt( Name, lua_tointeger( L, 3 ), Context );
		}
		else if( Var.m_Type == ConfigVar::EVT_Float )
		{
			ConfigManager::SetFloat( Name, lua_tonumber( L, 3 ), Context );
		}
		else
		{
			goto type_error;
		}
		break;
	default:
		goto type_error;
	}
	return 0;

type_error:
	return luaL_error( L, "invalid type for config var %s:%s", lua_tostring( L, 2 ),  lua_tostring( L, 1 ) );
}

/* Function list
 */
static const struct luaL_Reg Functions[] = {
	{ "hash", ConfigHash },
	{ "load", ConfigLoad },
	{ "loadstring", ConfigLoadstring },
	{ "save", ConfigSave },
	{ "getvar", ConfigGetvar },
	{ "setvar", ConfigSetvar },
 	{ NULL, NULL }
 };

int LuaLibraries::ConfigInit()
{
	LuaInterpreter *Lua = LuaInterpreter::GetInstance();
	return Lua->RegisterNamespace( "config", Functions );
}
