#include "core.h"
#include "hashedstring.h"
#include "luainterpreter.h"
#include "lualibraries.h"
#include "idatastream.h"
#include "dynamicmemorystream.h"
#include "packstream.h"
#include "modmanager.h"
#include "runtimeerror.h"

static int StreamCreate( lua_State *L )
{
	LuaInterpreter::PushUserdata<DynamicMemoryStream>( L );
	return 1;
}

static int StreamOpen( lua_State *L )
{
	const char *Filename = luaL_checkstring( L, 1 );
	try
	{
		LuaInterpreter::PushUserdata<PackStream>( L, Filename );
	}
	catch (RuntimeError Err)
	{
		lua_pushnil( L );
	}
	return 1;
}

/* Function list
 */
static const luaL_Reg Functions[] = {
	{ "create", StreamCreate },
	{ "open", StreamOpen },
 	{ NULL, NULL }
};

static int StreamWrite( lua_State *L )
{
	const DynamicMemoryStream *Stream = LuaInterpreter::GetUserdata<DynamicMemoryStream>( L, 1 );
	size_t DataSize = 0;
	const char *Data = luaL_checklstring( L, 2, &DataSize );
	Stream->Write( DataSize, Data );
	return 1;
}

static IDataStream *GetStream( lua_State *L, int Pos )
{
	IDataStream *Stream = LuaInterpreter::GetUserdata<DynamicMemoryStream>( L, Pos );
	if( !Stream )
	{
		Stream = LuaInterpreter::GetUserdata<PackStream>( L, Pos );
	}
	if( !Stream )
	{
		(void) luaL_argerror( L, Pos, "expected stream" );
		return NULL; // never gets here;
	}
	return Stream;
}

static int StreamRead( lua_State *L )
{
	const IDataStream *Stream = GetStream( L, 1 );
	if( Stream->EOS() )
	{
		lua_pushnil( L );
		return 1;
	}

	int Size = luaL_optinteger( L, 2, -1 );
	if( Size < 0 )
	{
		Size = Stream->Size() - Stream->GetPos();
	}
	char Buffer[ Size ];
	if( Stream->Read( Size, Buffer ) )
	{
		lua_pushlstring( L, Buffer, Size );
	}
	else
	{
		lua_pushnil( L );
	}
	return 1;
}

static int StreamSize( lua_State *L )
{
	const IDataStream *Stream = GetStream( L, 1 );
	lua_pushinteger( L, Stream->Size() );
	return 1;
}

static int StreamGetpos( lua_State *L )
{
	const IDataStream *Stream = GetStream( L, 1 );
	lua_pushinteger( L, Stream->GetPos() );
	return 1;
}

static int StreamSetpos( lua_State *L )
{
	const IDataStream *Stream = GetStream( L, 1 );
	int Pos = luaL_optinteger( L, 2, 0 );
	lua_pushboolean( L, Stream->SetPos( Pos ) );
	return 1;
}

static int StreamAtEof( lua_State *L )
{
	const IDataStream *Stream = GetStream( L, 1 );
	lua_pushboolean( L, Stream->EOS() );
	return 1;
}

static const luaL_Reg OMethods[] = {
	{ "write", StreamWrite },
	{ "read", StreamRead },
	{ "size", StreamSize },
	{ "getpos", StreamGetpos },
	{ "setpos", StreamSetpos },
	{ "ateof", StreamAtEof },
 	{ NULL, NULL }
};

static int OGc( lua_State *L )
{
	const DynamicMemoryStream *Stream = LuaInterpreter::GetUserdata<DynamicMemoryStream>( L, 1 );
	Stream->~DynamicMemoryStream();
	return 0;
}

static const luaL_Reg IMethods[] = {
	{ "read", StreamRead },
	{ "size", StreamSize },
	{ "getpos", StreamGetpos },
	{ "setpos", StreamSetpos },
	{ "ateof", StreamAtEof },
 	{ NULL, NULL }
};

static int IGc( lua_State *L )
{
	const PackStream *Stream = LuaInterpreter::GetUserdata<PackStream>( L, 1 );
	Stream->~PackStream();
	return 0;
}

int LuaLibraries::StreamInit()
{
	LuaInterpreter *Lua = LuaInterpreter::GetInstance();
	Lua->RegisterUsertype<DynamicMemoryStream>( OGc, NULL, OMethods );
	Lua->RegisterUsertype<PackStream>( IGc, NULL, IMethods );
	return Lua->RegisterNamespace( "stream", Functions );
}
