#include "core.h"
#include "hashedstring.h"
#include "luainterpreter.h"
#include "lualibraries.h"
#include "rawtexture.h"
#include "dynamicmemorystream.h"
#include "runtimeerror.h"

#include <string.h>

struct TextureObject {
	RawTexture *m_Texture;

	TextureObject()
	: m_Texture ( new RawTexture )
	{}

	~TextureObject()
	{
		delete m_Texture;
	}
};

static int TextureCreate( lua_State *L )
{
	uint Width = luaL_checkinteger( L, 1 );
	uint Height = luaL_checkinteger( L, 2 );
	TextureObject *Obj = LuaInterpreter::PushUserdata<TextureObject>( L );
	Obj->m_Texture->Initialize( Width, Height );
	return 1;
}

static int TextureLoad( lua_State *L )
{
	const char *Filename = luaL_checkstring( L, 1 );
	TextureObject *Obj = LuaInterpreter::PushUserdata<TextureObject>( L );
	int Top = lua_gettop( L );
	try
	{
		Obj->m_Texture->Initialize( Filename );
		// if the texture was loaded from a memory stream, that will still be present on  the stack
		lua_settop( L, Top );
	}
	catch (RuntimeError Err)
	{
		lua_settop( L, 1 );
		lua_pushnil( L );
	}
	return 1;
}

/* Function list
 */
static const luaL_Reg Functions[] = {
	{ "create", TextureCreate },
	{ "load", TextureLoad },
 	{ NULL, NULL }
};

static int TextureSave( lua_State *L )
{
	const TextureObject *Obj = LuaInterpreter::GetUserdata<TextureObject>( L, 1 );
	const char *Filename = luaL_checkstring( L, 2 );
	Obj->m_Texture->Save( Filename );
	return 0;
}

static int TextureSaveTo( lua_State *L )
{
	const TextureObject *Obj = LuaInterpreter::GetUserdata<TextureObject>( L, 1 );
	DynamicMemoryStream *Stream = LuaInterpreter::GetUserdata<DynamicMemoryStream>( L, 2 );
	if( Obj && Stream )
	{
		Obj->m_Texture->Save( *Stream );
	}
	return 0;
}

static int TextureCopyrect( lua_State *L )
{
	const TextureObject *From = LuaInterpreter::GetUserdata<TextureObject>( L, 1 );
	uint FromX = luaL_checkinteger( L, 2 );
	uint FromY = luaL_checkinteger( L, 3 );
	uint Width = luaL_checkinteger( L, 4 );
	uint Height = luaL_checkinteger( L, 5 );
	const TextureObject *To = LuaInterpreter::GetUserdata<TextureObject>( L, 6 );
	uint ToX = luaL_checkinteger( L, 7 );
	uint ToY = luaL_checkinteger( L, 8 );

	To->m_Texture->CopyRect( *From->m_Texture, FromX, FromY, Width, Height, ToX, ToY );

	return 0;
}

static int TextureSetPixel( lua_State *L )
{
	const TextureObject *Obj = LuaInterpreter::GetUserdata<TextureObject>( L, 1 );
	uint X = luaL_checkinteger( L, 2 );
	uint Y = luaL_checkinteger( L, 3 );
	uint R = luaL_checkinteger( L, 4 ) & 0xFF;
	uint G = luaL_checkinteger( L, 5 ) & 0xFF;
	uint B = luaL_checkinteger( L, 6 ) & 0xFF;
	uint A = luaL_checkinteger( L, 7 ) & 0xFF;
	Obj->m_Texture->SetPixel( X, Y, R, G, B, A );
	return 0;
}

static int TextureGetPixel( lua_State *L )
{
	const TextureObject *Obj = LuaInterpreter::GetUserdata<TextureObject>( L, 1 );
	uint X = luaL_checkinteger( L, 2 );
	uint Y = luaL_checkinteger( L, 3 );
	uint8 R, G, B, A;

	if( Obj->m_Texture->GetPixel( X, Y, R, G, B, A ) )
	{
		lua_pushinteger( L, R );
		lua_pushinteger( L, G );
		lua_pushinteger( L, B );
		lua_pushinteger( L, A );
		return 4;
	}
	return 0;
}

static const luaL_Reg Methods[] = {
	{ "save", TextureSave },
	{ "saveto", TextureSaveTo },
	{ "copyrect", TextureCopyrect },
	{ "setpixel", TextureSetPixel },
	{ "getpixel", TextureGetPixel },
 	{ NULL, NULL }
};

static int Gc( lua_State *L )
{
	TextureObject *Obj = LuaInterpreter::GetUserdata<TextureObject>( L, 1 );
	Obj->~TextureObject();
	return 0;
}

static int Index( lua_State *L )
{
	TextureObject *Obj = LuaInterpreter::GetUserdata<TextureObject>( L, 1 );
	const char *What = lua_tostring( L, 2 );
	if ( !strcmp( What, "sizex") )
	{
		lua_pushinteger( L, Obj->m_Texture->GetWidth() );
	}
	else if ( !strcmp( What, "sizey" ) )
	{
		lua_pushinteger( L, Obj->m_Texture->GetHeight() );
	}
	else
	{
		lua_pushnil( L );
	}
	return 1;
}

int LuaLibraries::TextureInit()
{
	LuaInterpreter *Lua = LuaInterpreter::GetInstance();
	Lua->RegisterUsertype<TextureObject>( Gc, Index, Methods );
	return Lua->RegisterNamespace( "texture", Functions );
}
