#include "core.h"
#include "filestream.h"
#include "hashedstring.h"
#include "simplestring.h"
#include "reversehash.h"
#include "array.h"
#include "hashreverser.h"
#include "printmanager.h"

static bool NextEntry( char *&start, char *&end )
{
	if ( *start ) {
		// skip empty lines
		while ( *start && *start <= ' ' )
		{
			++start;
		}

		// skip comments
		if ( *start == '#' )
		{
			while( *start && *start != '\r' && *start != '\n' )
			{
				++start;
			}
			return false;
		}

		end = start;

		while ( *end > ' ' ) {
			++end;
		}
		*end = '\0';

		return true;
	}
	return false;
}

HashReverser::HashReverser( const char *InitialStrings, const char *SequencePatterns )
: m_Sequences()
{
	// initialize ReverseHash buffer with an initial set of strings
	FileStream Strings( InitialStrings, FileStream::EFM_Read );
	uint32 StrSize = Strings.Size();
	char StrBuffer[ StrSize + 1 ];
	Strings.Read( StrSize, StrBuffer );
	StrBuffer[ StrSize ] = '\0';

	char *start = &StrBuffer[ 0 ], *end = start;
	while ( *start ) {
		if( !NextEntry( start, end ) )
		{
			continue;
		}

		SimpleString Name = start;
		HashedString NameHash = Name;
		ReverseHash::RegisterHash( NameHash, Name );

		start = end + 1;
	}

	// now read sequence patterns from  file
	FileStream Sequences( SequencePatterns, FileStream::EFM_Read );
	uint32 SeqSize = Sequences.Size();
	char SeqBuffer[ SeqSize + 1 ];
	Sequences.Read( SeqSize, SeqBuffer );
	SeqBuffer[ SeqSize ] = '\0';

	start = &SeqBuffer[ 0 ]; end = start;
	while ( *start ) {
		if( !NextEntry( start, end ) )
		{
			continue;
		}

		m_Sequences.PushBack( SequenceEntry( start ) );

		start = end + 1;
	}

}

HashReverser::~HashReverser()
{
}

SimpleString HashReverser::Reverse( HashedString Hash )
{
	SimpleString Reversed = ReverseHash::ReversedHash( Hash );
	if( Reversed == "" )
	{
		Reversed = CheckSequence( Hash );
	}
	if( Reversed == "" && Hash.GetHash() != 0 )
	{
		Reversed = SimpleString::PrintF("~%08X", Hash.GetHash() );
		ReverseHash::RegisterHash( Hash, Reversed );
	}
	return Reversed;
}

void HashReverser::RegisterHash( const HashedString& Hash, const SimpleString& String )
{
	ReverseHash::RegisterHash( Hash, String );
}

SimpleString HashReverser::CheckSequence( HashedString Hash )
{
	FOR_EACH_ARRAY( Test, m_Sequences, SequenceEntry )
	{
		SequenceEntry &Entry = Test.GetValue();
		if( Entry.CheckHash( Hash ) )
		{
			return Entry.GetLastValue();
		}
	}
	return "";
}
