#include "core.h"
#include "configmanager.h"
#include "filestream.h"
#include "reversehash.h"
#include "hashreverser.h"
#include "printmanager.h"
#include "cmdlineparser.h"
#include "file.h"
#include <string.h>
#include <stdio.h>
#include <math.h>

// Syntax: ConfigDecompiler.exe [-n names.txt] [-s sequences.txt] inconfig.config outconfig.ccf

bool IsName( const char *Value )
{
	if ( (*Value >= 'A' && *Value <= 'Z') || (*Value >= 'a' && *Value <= 'z') ) {
		return false;
	}
	Value++;
	while( *Value )
	{
		if ( *Value > ' ' ) {
			++Value;
		} else {
			break;
		}
	}
	return *Value == '\0';
}

void RehashValues( HashReverser &Reverser )
{
	ConfigManager *Config = ConfigManager::GetInstance();

	const ConfigManager::Vars& Contexts = ( *Config->GetContexts() );
	FOR_EACH_MAP( ContextIterator, Contexts, HashedString, ConfigManager::VarMap )
	{
		const ConfigManager::VarMap& Variables = ContextIterator.GetValue();
		FOR_EACH_MAP( VarIterator, Variables, HashedString, ConfigVar )
		{
			const ConfigVar& Value = VarIterator.GetValue();
			if( Value.m_Type == ConfigVar::EVT_String && IsName( Value.m_String ) )
			{
				SimpleString Name = Value.m_String;
				HashedString NameHash = Name;
				Reverser.RegisterHash( NameHash, Name );
			}
		}
	}
}

void DumpConfig( const char *CCFName, const IDataStream& Stream, HashReverser &Reverser )
{
	ConfigManager *Config = ConfigManager::GetInstance();

	Stream.PrintF( "# %s decompiled by mConfigDecompiler\n", CCFName );

	const ConfigManager::Vars& Contexts = ( *Config->GetContexts() );
	FOR_EACH_MAP( ContextIterator, Contexts, HashedString, ConfigManager::VarMap )
	{
		const HashedString& ContextKey = ContextIterator.GetKey();
		const ConfigManager::VarMap& Variables = ContextIterator.GetValue();
		SimpleString ContextName = Reverser.Reverse( ContextKey );
		if( ContextName != "" )
		{
			Stream.PrintF( "\n[%s]\n", ContextName.CStr() );	// Hashed context name
		}

		FOR_EACH_MAP( VarIterator, Variables, HashedString, ConfigVar )
		{
			const HashedString& Key = VarIterator.GetKey();
			const ConfigVar& Value = VarIterator.GetValue();
			SimpleString KeyName = Reverser.Reverse( Key );
			const char *Indent = ( ContextName == "" ? "" : "\t" );

			switch( Value.m_Type )
			{
			case ConfigVar::EVT_Bool:
				Stream.PrintF( "%s%s = %s\n", Indent, KeyName.CStr(), Value.m_Bool ? "true" : "false" );
				break;
			case ConfigVar::EVT_Int:
				Stream.PrintF( "%s%s = %d\n", Indent, KeyName.CStr(), Value.m_Int );
				break;
			case ConfigVar::EVT_Float:
				if ( Value.m_Float == floor( Value.m_Float ) )
				{
					Stream.PrintF( "%s%s = %g.0\n", Indent, KeyName.CStr(), Value.m_Float );
				}
				else
				{
					Stream.PrintF( "%s%s = %.7g\n", Indent, KeyName.CStr(), Value.m_Float );
				}
				break;
			case ConfigVar::EVT_String:
				Stream.PrintF( "%s%s = \"%s\"\n", Indent, KeyName.CStr(), Value.m_String );
				break;
			default:
				WARNDESC( "Bad config var type" );
				break;
			}
		}
	}
}

CmdlineParser ParseCommandline( int Argc, char *Argv[] )
{
	SimpleString NamesDefault = FileUtil::GetExecutableDirectory() + "/names.txt";
	SimpleString SeqDefault = FileUtil::GetExecutableDirectory() + "/sequences.txt";

	CmdlineParser Opts( Argc, Argv );
	Opts.AddOption( 'n', "names", "file with the list of initial names", true, NamesDefault.CStr() );
	Opts.AddOption( 's', "sequences", "file with the list of sequence patterns", true, SeqDefault.CStr() );
	Opts.AddPositionals( 2, "<infile> <outfile>", "input and output files" );
	if ( !Opts.Parse() )
	{
		puts( Opts.Usage().CStr() );
		exit( 1 );
	}
	return Opts;
}

int main( int Argc, char* Argv[] )
{
	CmdlineParser Opts = ParseCommandline( Argc, Argv );

	ReverseHash::Initialize();

	HashReverser Reverser = HashReverser( Opts["names"], Opts["sequences"] );
	ConfigManager::Load( FileStream( Opts[1], FileStream::EFM_Read ) );
	RehashValues( Reverser );
	DumpConfig( Opts[1], FileStream( Opts[2], FileStream::EFM_Write ), Reverser );

	ReverseHash::ShutDown();
	return 0;
}
